'From OLPC2.0 of ''24 October 2006'' [latest update: #1349] on 22 May 2007 at 9:49:53 am'!

!TTCFont methodsFor: 'friend' stamp: 'bf 6/13/2006 18:47'!
cache
	cache size = (self maxAscii + 1) ifFalse:[self recreateCache].
	^cache! !

!TTCFont methodsFor: 'friend' stamp: 'bf 6/13/2006 18:11'!
recreateCache

	cache _ WeakArray new: self maxAscii + 1.! !

!TTCFont methodsFor: 'public' stamp: 'bf 6/13/2006 18:25'!
widthOf: aCharacter

	"This method cannot use #formOf: because formOf: discriminates the color and causes unnecessary bitmap creation."

	| f assoc |
	(self hasGlyphOf: aCharacter) ifFalse: [
		fallbackFont ifNotNil: [^ fallbackFont widthOf: aCharacter].
		^ 1
	].
	assoc _ self cache at: (aCharacter charCode + 1).
	assoc ifNotNil: [
		^ assoc value width
	].

	f _ self computeForm: aCharacter.
	self at: aCharacter charCode put: f.
	^ f width.
! !

!TTCFont methodsFor: 'private' stamp: 'bf 6/13/2006 18:25'!
formOf: char

	| f assoc code |
	(self hasGlyphOf: char)
		ifFalse: [^ self fallbackFont formOf: char].

	code _ char charCode.
	assoc _ self cache at: (code + 1).
	assoc ifNotNil: [
		(assoc key = foregroundColor) ifTrue: [
			^ assoc value.
		].
	].

	f _ self computeForm: code.
	self at: code put: f.
	^ f.
! !

!TTCFont methodsFor: 'private' stamp: 'bf 6/13/2006 18:13'!
hasGlyphOf: aCharacter

	^ aCharacter charCode <= self maxAscii
! !


!TTFontDescription methodsFor: 'accessing' stamp: 'bf 6/12/2006 18:14'!
size

	^ glyphTable size - 1
! !


!TTFontReader methodsFor: 'processing' stamp: 'bf 6/12/2006 18:00'!
processCharMap: assoc
	"Process the given character map"

	| charTable glyph cmap |
	cmap _ assoc value.

	assoc key = 0 ifTrue: "Unicode table"
		[charTable := SparseLargeTable new: cmap size
			chunkSize: 256 arrayClass: Array base: 1
			defaultValue: glyphs first.
		1 to: charTable size do:
			[:i |
			glyph _ glyphs at: (cmap at: i) + 1 ifAbsent: [glyphs first].
			charTable at: i put: glyph].
		charTable zapDefaultOnlyEntries.
		^charTable].

	charTable _ Array new: 256 withAll: glyphs first. "Initialize with default glyph"

	assoc key = 1 ifTrue: "Mac encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph _ glyphs at: (cmap at: i) + 1.
			charTable at: (self macToWin: i) put: glyph]].

	assoc key = 3 ifTrue: "Win encoded table"
		[1 to: (cmap size min: charTable size) do:
			[:i |
			glyph _ glyphs at: (cmap at: i) + 1.
			charTable at: i put: glyph]].

	^ charTable! !

!TTFontReader methodsFor: 'processing' stamp: 'bf 6/12/2006 16:28'!
processCharacterMappingTable: entry
	"Read the font's character to glyph index mapping table.
	If an appropriate mapping can be found then return an association
	with the format identifier and the contents of the table"
	| copy initialOffset nSubTables pID sID offset cmap assoc |
	initialOffset _ entry offset.
	entry skip: 2. "Skip table version"
	nSubTables _ entry nextUShort.
	1 to: nSubTables do:[:i|
		pID _ entry nextUShort.
		sID _ entry nextUShort.
		offset _ entry nextULong.
		"Check if this is either a Unicode (0), Macintosh (1),
		or a Windows (3) encoded table"
		(#(0 1 3) includes: pID) ifTrue:[
			"Go to the beginning of the table"
			copy _ entry copy.
			copy offset: initialOffset + offset.
			cmap _ self decodeCmapFmtTable: copy.
			(pID = 0 and: [cmap notNil]) "Prefer Unicode encoding over everything else"
				ifTrue: [^ pID -> cmap].
			assoc _ pID -> cmap. "Keep it in case we don't find a better table"
		].
	].
	^assoc! !

!TTFontReader methodsFor: 'private' stamp: 'bf 6/12/2006 17:14'!
decodeCmapFmtTable: entry
	| cmapFmt length cmap firstCode entryCount segCount segments offset code |
	cmapFmt _ entry nextUShort.
	length _ entry nextUShort.
	entry skip: 2. "skip version"

	cmapFmt = 0 ifTrue: "byte encoded table"
		[length _ length - 6. 		"should be always 256"
		length <= 0 ifTrue: [^ nil].	"but sometimes, this table is empty"
		cmap _ Array new: length.
		entry nextBytes: length into: cmap startingAt: entry offset.
		^ cmap].

	cmapFmt = 4 ifTrue: "segment mapping to deltavalues"
		[segCount _ entry nextUShort // 2.
		entry skip: 6. "skip searchRange, entrySelector, rangeShift"
		segments _ Array new: segCount.
		segments _ (1 to: segCount) collect: [:e | Array new: 4].
		1 to: segCount do: [:i | (segments at: i) at: 2 put: entry nextUShort]. "endCount"
		entry skip: 2. "skip reservedPad"
		1 to: segCount do: [:i | (segments at: i) at: 1 put: entry nextUShort]. "startCount"
		1 to: segCount do: [:i | (segments at: i) at: 3 put: entry nextShort]. "idDelta"
		offset _ entry offset.
		1 to: segCount do: [:i | (segments at: i) at: 4 put: entry nextUShort]. "idRangeOffset"
		entryCount _ segments inject: 0 into: [:max :seg | max max: seg second].
		cmap _ Array new: entryCount+1 withAll: 0..
		segments withIndexDo:
			[:seg :si |
			seg first to: seg second do:
				[:i |
				seg last > 0 ifTrue:
					["offset to glypthIdArray - this is really C-magic!!"
					entry offset: i - seg first - 1 * 2 + seg last + si + si + offset. 
					code _ entry nextUShort.
					code > 0 ifTrue: [code _ code + seg third]]
				ifFalse:
					["simple offset"
					code _ i + seg third].
				cmap at: i + 1 put: code]].
		^ cmap].

	cmapFmt = 6 ifTrue: "trimmed table"
		[firstCode _ entry nextUShort.
		entryCount _ entry nextUShort.
		cmap _ Array new: entryCount + firstCode withAll: 0.
		entryCount timesRepeat:
			[cmap at: (firstCode _ firstCode + 1) put: entry nextUShort].
		^ cmap].
	^ nil! !

